import numpy as np
from scipy.linalg import eigh  # can also solve generalized EVD (compared to numpy.linalg.eigh)
from scipy.linalg import sqrtm, block_diag

from .helpers import vectorize_datasets, check_zero_mean, make_ccvs_unit_variance


def mcca(X, algorithm='genvar', max_iter=1000, eps=0.0001, verbose=False, C_xx=None, W_init=None):
    """
    Perform mCCA on datasets X^[k].

    Info
    ----
    N: number of components, T: number of samples, K: number of datasets

    Parameters
    ----------
    X : np.ndarray
        Datasets of dimensions N x T x K, where X[:,:,k] corresponds to X^[k]

    algorithm : str, optional
        mCCA algorithm: possible options are: 'sumcor', 'maxvar', 'mivar', 'ssqcor', 'genvar'

    max_iter : int
        Maximum number of iterations for the numerical algorithms (genvar and ssqcor)

    eps : float, optional
        Threshold value for convergence. If change of theta parameter is smaller than eps, optimization will stop.

    verbose : bool, optional
        If True, print after how many iterations the algorithm stopped for each SCV

    C_xx : np.ndarray
        True covariance matrix of the datasets (not estimated using samples). If C_xx is provided, the measured
        performance is for the infinite sample case.

    Returns
    -------
    M : np.ndarray
        Transformation matrices of dimension N x N x K, such that E^[k] = (M^[k])^T X^[k]

    Epsilon : np.ndarray
        Canonical variables of dimension N x T x K

    """

    if algorithm == 'sumcor':
        M, Epsilon = mcca_sumcor_nielsen(X, C_xx)
    elif algorithm == 'maxvar':
        M, Epsilon = mcca_maxvar_minvar_kettenring(X, 'maxvar', C_xx)
    elif algorithm == 'minvar':
        M, Epsilon = mcca_maxvar_minvar_kettenring(X, 'minvar', C_xx)
    elif algorithm == 'ssqcor':
        M, Epsilon = mcca_ssqcor_genvar_kettenring(X, 'ssqcor', max_iter, eps, verbose, C_xx, W_init)
    elif algorithm == 'genvar':
        M, Epsilon = mcca_ssqcor_genvar_kettenring(X, 'genvar', max_iter, eps, verbose, C_xx, W_init)
    else:
        raise AssertionError("'algorithm' must be 'sumcor', 'maxvar', 'minvar', 'ssqcor', or 'genvar'.")

    return M, Epsilon


def mcca_sumcor_nielsen(X, C_xx=None):
    """
    Implementation of mCCA-sumcor according to
    Nielsen, Allan Aasbjerg. "Multiset canonical correlations analysis and multispectral, truly multitemporal remote
    sensing data." IEEE transactions on image processing 11.3 (2002): 293-305.


    Info
    ----
    N: number of components, T: number of samples, K: number of datasets

    Parameters
    ----------
    X : np.ndarray
        Datasets of dimensions N x T x K, where X[:,:,k] corresponds to X^[k]

    Returns
    -------
    M : np.ndarray
        Transformation matrices of dimension N x N x K, such that E^[k] = (M^[k])^T X^[k]

    Epsilon : np.ndarray
        Canonical variables of dimension N x T x K

    """

    N, T, K = X.shape

    # make sure data is zero-mean
    check_zero_mean(X)

    if C_xx is None:
        # stack datasets -> [x^[1]^T, ..., x^[K]^T]^T
        X_concat = vectorize_datasets(X)
        C_xx = np.cov(X_concat, ddof=0)

    # cut diagonal blocks of C and store them in D_xx
    D_xx = np.zeros_like(C_xx)
    for k in range(K):
        D_xx[k * N:(k + 1) * N, k * N:(k + 1) * N] = C_xx[k * N:(k + 1) * N, k * N:(k + 1) * N]

    # solve GEVD C w_n = lambda_n D_xx w_n
    eigvals, eigvecs = eigh(C_xx, D_xx)

    # take only N largest eigenvalues and corresponding eigenvectors
    eigvecs = eigvecs[:, ::-1]  # sort ascending
    M_tilde = eigvecs[:, 0:N]  # eigenvectors corresponding to N largest EVs

    # match elements of M to M^[k]
    M = np.zeros((N, N, K))
    for k in range(K):
        M[:, :, k] = M_tilde[k * N:(k + 1) * N, :]

    # calculate canonical variates
    Epsilon = np.zeros_like(X)
    for k in range(K):
        Epsilon[:, :, k] = M[:, :, k].T @ X[:, :, k]

    # normalize canonical variables to unit variance (and save scalings in transformation matrices)
    M, Epsilon = make_ccvs_unit_variance(M, Epsilon)

    return M, Epsilon


def mcca_maxvar_minvar_kettenring(X, algorithm, C_xx=None):
    """
    Implementation of mCCA-maxvar and mCCA-minvar according to
    Kettenring, J. R. (1971). Canonical analysis of several sets of variables. Biometrika, 58(3), 433-451.

    Info
    ----
    N: number of components, T: number of samples, K: number of datasets

    Parameters
    ----------
    X : np.ndarray
        Datasets of dimensions N x T x K, where X[:,:,k] corresponds to X^[k]

    algorithm: str
        'maxvar' or 'minvar'

    Returns
    -------
    M : np.ndarray
        Transformation matrices of dimension N x N x K, such that E^[k] = (M^[k])^T X^[k]

    Epsilon : np.ndarray
        Canonical variables of dimension N x T x K

    """

    N, T, K = X.shape

    # make sure data is zero-mean
    check_zero_mean(X)

    if C_xx is None:
        # stack datasets -> [x^[1]^T, ..., x^[K]^T]^T
        X_concat = vectorize_datasets(X)
        C_xx = np.cov(X_concat, ddof=0)

        # whiten x^[k] -> y^[k] (using Mahalanobis whitening)
        Y = np.zeros_like(X)
        for k in range(K):
            # ddof=0 means dividing by T, not T-1
            Y[:, :, k] = np.linalg.inv(sqrtm(C_xx[k * N:(k + 1) * N, k * N:(k + 1) * N])) @ X[:, :, k]

        # concatenate whitened datasets vertically
        Y_concat = vectorize_datasets(Y)

        # calculate C_yy = E[y y^T]
        C_yy = np.cov(Y_concat, ddof=0)

    else:
        # cut diagonal blocks of C_xx and store them in D_xx
        D_xx = np.zeros_like(C_xx)
        for k in range(K):
            D_xx[k * N:(k + 1) * N, k * N:(k + 1) * N] = C_xx[k * N:(k + 1) * N, k * N:(k + 1) * N]

        C_yy = np.linalg.inv(sqrtm(D_xx)) @ C_xx @ np.linalg.inv(sqrtm(D_xx))

    # calculate the canonical variates in a deflationary way for each stage n = 1...N
    V = np.zeros((N * K, N))
    V_tilde = np.zeros((N, N, K))
    for n in range(N):

        if n == 0:
            # H_n for the first stage is the identity matrix
            H_n = np.eye(N * K)
        else:
            # update H_n matrix

            # stack K partitions of V_tilde as diagonal blocks of V_tilde_n1 (of dimensions NK x (n-1)K )
            # Kettenring eq. 9.9: D_c = diag{1_C^(s), ..., m_C^{(s)}, j_C^(s) = {j_b^(1), ..., j_b^(s-1)}
            V_list = [V_tilde[:, 0:n, k] for k in range(K)]  # V_tilde is defined later
            V_tilde_n1 = block_diag(*V_list)

            # update H_n
            H_n = np.eye(N * K) - V_tilde_n1 @ np.linalg.inv(V_tilde_n1.T @ V_tilde_n1) @ V_tilde_n1.T

        # calc v_n as first eigvec of EVD(H_n R H_n), where H_1 = I
        eigval, eigvec = eigh(H_n @ C_yy @ H_n)
        # eigvals and eigvecs should be sorted in descending order
        eigvec = eigvec[:, ::-1]

        if algorithm == 'maxvar':
            # eigenvector corresponding to largest eigenvalue
            V[:, n] = eigvec[:, 0]
        elif algorithm == 'minvar':
            # eigenvector corresponding to smallest non-zero eigenvalue
            V[:, n] = eigvec[:, N * K - n * K - 1]
        else:
            raise ValueError("'algorithm' must be 'maxvar' or 'minvar'")

        # normalize each v_n^[k] -> w_n[k] has unit norm
        for k in range(K):
            v_n_k = V[k * N:(k + 1) * N, n]
            V_tilde[:, n, k] = v_n_k / np.linalg.norm(v_n_k)

    # now all m_n^[k] are calculated (Kettenring: j_b^(s), where k=j and n=s)

    # calculating demixing matrices to multiply with x^[k] instead of y^[k]
    M = np.zeros_like(V_tilde)
    for k in range(K):
        M[:, :, k] = np.linalg.inv(sqrtm(C_xx[k * N:(k + 1) * N, k * N:(k + 1) * N])) @ V_tilde[:, :, k]

    # calculate canonical variates (they already have unit variance)
    Epsilon = np.zeros_like(X)
    for k in range(K):
        Epsilon[:, :, k] = M[:, :, k].T @ X[:, :, k]

    return M, Epsilon


def mcca_ssqcor_genvar_kettenring(X, algorithm, max_iter=1000, eps=0.0001, verbose=False, C_xx=None,
                                  W_init=None):
    """
    Implementation of mCCA-ssqcor and mCCA-genvar according to
    Kettenring, J. R. (1971). Canonical analysis of several sets of variables. Biometrika, 58(3), 433-451.

    Info
    ----
    N: number of components, T: number of samples, K: number of datasets

    Parameters
    ----------
    X : np.ndarray
        Datasets of dimensions N x T x K, where X[:,:,k] corresponds to X^[k]

    max_iter : int, optional
        Maximum number of iterations before stopping the optimization

    eps : float, optional
        Threshold value for convergence. If change of theta parameter is smaller than eps, optimization will stop.

    verbose : bool, optional
        If True, print after how many iterations the algorithm stopped for each SCV

    Returns
    -------
    M : np.ndarray
        Transformation matrices of dimension N x N x K, such that E^[k] = (M^[k])^T X^[k]

    Epsilon : np.ndarray
        Canonical variables of dimension N x T x K

    """

    N, T, K = X.shape

    # make sure data is zero-mean
    check_zero_mean(X)

    if C_xx is None:
        # stack datasets -> [x^[1]^T, ..., x^[K]^T]^T
        X_concat = vectorize_datasets(X)
        C_xx = np.cov(X_concat, ddof=0)

        # whiten x^[k] -> y^[k] (using Mahalanobis whitening)
        Y = np.zeros_like(X)
        for k in range(K):
            # ddof=0 means dividing by T, not T-1
            Y[:, :, k] = np.linalg.inv(sqrtm(C_xx[k * N:(k + 1) * N, k * N:(k + 1) * N])) @ X[:, :, k]

        # concatenate whitened datasets vertically
        Y_concat = vectorize_datasets(Y)

        # calculate C_yy = E[y y^T]
        C_yy = np.cov(Y_concat, ddof=0)

    else:
        # cut diagonal blocks of C_xx and store them in D_xx
        D_xx = np.zeros_like(C_xx)
        for k in range(K):
            D_xx[k * N:(k + 1) * N, k * N:(k + 1) * N] = C_xx[k * N:(k + 1) * N, k * N:(k + 1) * N]

        C_yy = np.linalg.inv(sqrtm(D_xx)) @ C_xx @ np.linalg.inv(sqrtm(D_xx))

    if W_init is not None:  # given W_init is for X, but calculations are for Y -> multiply by whitening matrix
        V = np.zeros((N, N, K))
        for k in range(K):
            # ddof=0 means dividing by T, not T-1
            V[:, :, k] = sqrtm(C_xx[k * N:(k + 1) * N, k * N:(k + 1) * N]) @ W_init[:, :, k].T
        pass
    else:
        # initialize orthogonal transformation matrices (that would transform Y)
        V = np.random.randn(N, N, K)
        for k in range(K):
            V[:, :, k] = np.linalg.solve(sqrtm(V[:, :, k] @ V[:, :, k].T), V[:, :, k])

    for n in range(N):
        theta_n = np.zeros((K, max_iter))
        H_n = np.zeros((N, N, K))
        for k in range(K):
            if n == 0:
                H_n[:, :, k] = np.eye(N)  # A_j in eq. (12.5c) for the first CCV
                if algorithm == 'genvar':
                    # init R_n
                    V_n = block_diag(*V[:, n, :].T).T
                    R_n = V_n.T @ C_yy @ V_n  # K x K covariance matrix of nth CCV
            else:
                V_n_1_k = V[:, 0:n, k]  # C_j on p.445
                H_n[:, :, k] = np.eye(N) - V_n_1_k @ np.linalg.inv(
                    V_n_1_k.T @ V_n_1_k) @ V_n_1_k.T  # A_j in eq. (12.5c)

        for iter in range(1, max_iter):
            for k in range(K):
                # [ C_yy^[k,1] m_n^[1], ..., C_yy^[k,k-1] m_n^[k-1], C_yy^[k,k+1] m_n^[k+1], ..., C_yy^[k,K] m_n^[K] ]
                N_n_k = []
                for l in range(K):
                    if l != k:
                        N_n_k.append(C_yy[k * N: (k + 1) * N, l * N: (l + 1) * N] @ V[:, n, l])
                N_n_k = np.array(N_n_k).T  # called N_j in Kettenring's paper p.445, his j is our k

                if algorithm == 'ssqcor':
                    F_n_k = N_n_k @ N_n_k.T  # P_j on p.445
                elif algorithm == 'genvar':
                    # delete kth row and kth column of nth CCV covariance matrix R_n
                    R_n_minus_k = np.delete(np.delete(R_n, k, 0), k, 1)  # K-1 x K-1 matrix Phi_j on p.445, his j=our k
                    F_n_k = N_n_k @ np.linalg.inv(R_n_minus_k) @ N_n_k.T  # Q_j on p.445
                else:
                    raise ValueError("'algorithm' must be 'ssqcor' or 'genvar'")

                # we can perform EVD of H_n F_n H_n (which is faster), as we are just interested in leading eigenvector
                eigval, eigvec = np.linalg.eigh(H_n[:, :, k] @ F_n_k @ H_n[:, :, k])
                # eigvals and eigvecs should be sorted in descending order
                eigval = eigval[::-1]
                eigvec = eigvec[:, ::-1]
                V[:, n, k] = eigvec[:, 0]
                theta_n[k, iter] = eigval[0]

                if algorithm == 'genvar':
                    # update R_n (K x K covariance matrix of nth CCV)
                    V_n = block_diag(*V[:, n, :].T).T
                    R_n = V_n.T @ C_yy @ V_n  # K x K covariance matrix of nth CCV

                    # this would be much slower
                    # for k1 in range(K):
                    #     for k2 in range(K):
                    #         R_n[k1, k2] = V[:, n, k1].T @ C_yy[k1 * N: (k1 + 1) * N, k2 * N: (k2 + 1) * N] @ V[:, n, k2]

            if np.sum(np.abs(theta_n[:, iter] - theta_n[:, iter - 1])) < eps:  # eq. (12.9)
                if verbose:
                    print(f'Stopping for the {n}th CCV after {iter} iterations')
                break

    # find transformation matrices for X^[k] instead of Y^[k]
    M = np.zeros((N, N, K))
    for k in range(K):
        M[:, :, k] = np.linalg.inv(sqrtm(C_xx[k * N:(k + 1) * N, k * N:(k + 1) * N])) @ V[:, :, k]

    # calculate canonical variates (they already have unit variance)
    Epsilon = np.zeros_like(X)
    for k in range(K):
        Epsilon[:, :, k] = M[:, :, k].T @ X[:, :, k]

    return M, Epsilon




import numpy as np
from scipy.stats import pearsonr


def vectorize_datasets(X):
    """
    Concatenate datasets vertically.

    Parameters
    ----------
    X : np.ndarray
        data of dimensions N x T x K

    Returns
    -------
    X_concat : np.ndarray
        stacked datasets of dimensions NK x T

    """

    N, T, K = X.shape
    # stack datasets -> [x^[1]^T, ..., x^[K]^T]^T
    X_concat = np.reshape(np.moveaxis(X, [0, 1, 2], [0, 2, 1]), (N * K, T), 'F')

    return X_concat


def check_zero_mean(X):
    N, T, K = X.shape
    # make sure data is zero-mean
    for k in range(K):
        np.testing.assert_almost_equal(np.mean(X[:, :, k], axis=1), 0)


def make_ccvs_unit_variance(M, Epsilon):
    # make Epsilon unit-variance and write std in W
    M_unit_var = np.zeros_like(M)
    Epsilon_unit_var = np.zeros_like(Epsilon)
    for k in range(Epsilon.shape[2]):
        std = np.std(Epsilon[:, :, k], axis=1, keepdims=True)
        Epsilon_unit_var[:, :, k] = Epsilon[:, :, k] / std
        M_unit_var[:, :, k] = M[:, :, k] / std.T
    return M_unit_var, Epsilon_unit_var


def calculate_ccv_covariance_matrices(Epsilon):
    N, _, K = Epsilon.shape
    ccv_cov = np.zeros((K, K, N))
    for n in range(N):
        ccv_cov[:, :, n] = np.cov(Epsilon[n, :, :].T, ddof=0)  # ddof=0 means dividing by N
    return ccv_cov


def calculate_eigenvalues_from_ccv_covariance_matrices(ccv_cov):
    K, _, N = ccv_cov.shape
    Lambda = np.zeros((N, K))
    for n in range(N):
        Lambda[n, :] = np.linalg.eigh(ccv_cov[:, :, n])[0]
    return Lambda


def calculate_avg_abs_pearson_correlation_coefficient(S_1, S_2):
    # calculate the Pearson correlation coefficient between the sources in S_1 and S_2 for all permutations
    # return the highest correlation coefficient, which is for the best permutation
    '''

    Parameters
    ----------
    S_1 : np.ndarray
        sources of dimension N x T x K
    S_2 : np.ndarray
        sources of dimension N x T x K

    Returns
    -------
    avg_pearson : float
        average value of the absolute Pearson correlation coefficient
    '''

    N, T, K = S_1.shape

    if S_1.shape != S_2.shape:
        raise AssertionError("'S_1' and 'S_2' must have the same shape.")

    pearsoncoefficient = np.zeros((N, N))
    for n in range(N):
        for m in range(N):
            # avg abs correlation coefficient for nth and mth source component in all K datasets
            pearsoncoefficient[n, m] = 0
            for k in range(K):
                pearsoncoefficient[n, m] += np.abs(pearsonr(S_1[n, :, k], S_2[m, :, k]).statistic)
            pearsoncoefficient[n, m] /= K

    # find the indices of the best permutation
    permutation = np.argmax(pearsoncoefficient, axis=1)

    # if column indices (elements in permutation) are not unique, find other match for duplicates
    if len(np.unique(permutation)) < N:

        # find duplicate column indices
        seen = set()
        dupes = set()
        for col in permutation:
            if col in seen:
                dupes.add(col)
            else:
                seen.add(col)

        # find corresponding row indices
        for col in dupes:
            row = np.where(permutation == col)[0]
            duplicate_pearson = pearsoncoefficient[row, col]
            # find row indices from largest to smallest Pearson correlation coefficient for that column
            sorted_row_idx = np.argsort(-duplicate_pearson)
            # row[sorted_row_idx[0]] has maximum pearson value
            # find new column idx for remaining row indices
            for row_idx in row[sorted_row_idx[1:]]:
                # find col indices from largest to smallest pearson correlation coefficients in that row
                sorted_col_idx = np.argsort(-pearsoncoefficient[row_idx, :])
                # update column idx (element in permutation) for that row
                for col_idx in range(1, len(sorted_col_idx)):
                    # check if second-largest element is not already a match for another row
                    if sorted_col_idx[col_idx] not in permutation:
                        # replace the element in permutation:
                        permutation[row_idx] = sorted_col_idx[col_idx]
                        break
                    else:
                        # wait for the next element
                        pass

    # avg correlation for this permutation
    avg_pearson = np.mean(pearsoncoefficient[np.arange(N), permutation])

    return avg_pearson



import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
import pandas as pd

from .helpers import calculate_eigenvalues_from_ccv_covariance_matrices


def plot_eigenvalues(scv_cov, title=None, show=True, filename=None):
    Lambda = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov)
    Lambda = Lambda[:, ::-1]  # sort descending

    plt.figure(figsize=(6, 3.5))
    for n in range(scv_cov.shape[2]):
        plt.plot(np.arange(1, Lambda.shape[1] + 1), Lambda[n, :], 'D:', markersize=2.5, lw=1,
                 label=r'$\mathbf{\lambda}_{' + f'{n + 1}' + r'}$')
        plt.xticks(np.arange(0, Lambda.shape[1] + 1, 5))
    plt.legend()

    plt.title(title)
    plt.tight_layout()
    if filename is not None:
        plt.savefig(filename, dpi=500)
    if show:
        plt.show()


def write_results_in_latex_table(K, n_montecarlo):
    results = np.load(Path(Path(__file__).parent.parent, f'simulation_results/K_{K}/K_{K}.npy'),
                      allow_pickle=True).item()

    # create pandas dataframe from results
    scenarios = list(results.keys())
    algorithms = list(results[scenarios[0]].keys())

    # store mean and std in array (as string?)
    joint_isi_mean_std_array = np.zeros((len(algorithms), len(scenarios)), dtype=object)
    runtime_mean_std_array = np.zeros((len(algorithms), len(scenarios)), dtype=object)
    for scenario_idx, scenario in enumerate(scenarios):
        for algorithm_idx, algorithm in enumerate(algorithms):
            joint_isi = results[scenario][algorithm]['joint_isi']
            joint_isi_mean_std_array[
                algorithm_idx, scenario_idx] = f'${np.round(np.mean(joint_isi), 2)} \pm {np.round(np.std(joint_isi), 2)}$'

            runtime = results[scenario][algorithm]['runtime']
            runtime_mean_std_array[
                algorithm_idx, scenario_idx] = f'${np.round(np.mean(runtime), 2)} \pm {np.round(np.std(runtime), 2)}$'

    table_headings = ['same $\lambda_{\mathrm{max}}$', 'same $\lambda_{\mathrm{min}}$', r'same $\vect{\lambda}$',
                      r'$\mat{C}_n = \mat{P}_n\tran \mat{C}_1 \mat{P}_n$',
                      '$\mat{C}_n = \mat{D}_n \mat{C}_1 \mat{D}_n$',
                      'rank $R=1$', 'rank $R=K$']
    joint_isi_df = pd.DataFrame(joint_isi_mean_std_array, columns=table_headings, index=algorithms)
    joint_isi_df.to_latex()
    runtime_df = pd.DataFrame(runtime_mean_std_array, columns=table_headings, index=algorithms)

    joint_isi_df.to_latex(Path(Path(__file__).parent.parent, f'simulation_results/joint_isi_K_{K}.tex'),
                          caption=r'joint ISI value (lower is better) for \underline{'
                                  f'$K={K}$'
                                  '} datasets, averaged across '
                                  f'{n_montecarlo} runs. '
                                  r'The sumcor algorithm is according to Nielsen \cite{Nielsen2002}, '
                                  r'the other algorithms are according to Kettenring \cite{Kettenring1971}.',
                          label='tab:jointisiresults',
                          position='!htb')

    runtime_df.to_latex(Path(Path(__file__).parent.parent, f'simulation_results/runtime_K_{K}.tex'),
                        caption=r'runtime in seconds (lower is better) for \underline{'
                                f'$K={K}$'
                                '} datasets, averaged across '
                                f'{n_montecarlo} runs.',
                        label='tab:runtimeresults',
                        position='!htb')



import numpy as np
import matplot2tikz
from pathlib import Path
import matplotlib.pyplot as plt

from .helpers import calculate_eigenvalues_from_ccv_covariance_matrices


def plot_identification_conditions_for_paper(folder, n_montecarlo, save=False):
    results_violations = np.load(Path(Path(__file__).parent.parent, f'simulation_results/{folder}/violations.npy'),
                                 allow_pickle=True).item()

    # store violation results for each algorithm
    scenarios_violations = ['same_eigenvalues_same_eigenvectors',
                            'same_eigenvalues_different_eigenvectors',
                            'different_lambda_min',
                            'different_lambda_max']  # dict keys would be in wrong order
    scenario_labels_violations = ['Exp. A', 'Exp. B', 'Exp. C', 'Exp. D']
    n_scenarios_violations = len(scenario_labels_violations)
    algorithms = ['sumcor', 'maxvar', 'minvar', 'genvar', 'ssqcor']
    joint_isi_per_algorithm_violations = {algorithm: np.zeros((n_scenarios_violations, n_montecarlo)) for algorithm
                                          in algorithms}

    for scenario_idx, scenario in enumerate(scenarios_violations):
        for algorithm_idx, algorithm in enumerate(algorithms):
            joint_isi_per_algorithm_violations[algorithm][scenario_idx, :] = results_violations[scenario][algorithm][
                'joint_isi']

    # plot joint ISI for violations using true and estimated covariance matrices in one figure
    fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(9, 2.5))

    colors = ['C0', 'C3', 'C2', 'C1', 'C4']

    # errorbar
    axes.axhline(y=0.05, color='tab:gray', linestyle=':', linewidth=1.1)
    for (idx, algorithm) in enumerate(algorithms):
        axes.errorbar(np.arange(idx, n_scenarios_violations * (len(algorithms) + 2), len(algorithms) + 2),
                      np.mean(joint_isi_per_algorithm_violations[algorithm], axis=1),
                      np.std(joint_isi_per_algorithm_violations[algorithm], axis=1),
                      color=colors[idx], linestyle='', fmt='D', markersize=3, capsize=2, lw=1.1,
                      label=f'{algorithm}')
    axes.set_xticks(np.arange(2, n_scenarios_violations * (len(algorithms) + 2), len(algorithms) + 2),
                    scenario_labels_violations, fontsize=12)
    axes.set_ylim([-0.05, 1.05])
    axes.set_yticks([0, 0.5, 1])
    axes.set_yticklabels([0, 0.5, 1], fontsize=12)
    axes.set_ylabel('jISI', fontsize=12)

    plt.legend(loc='center left', bbox_to_anchor=(1.05, 0.5))

    plt.tight_layout()
    plt.subplots_adjust(wspace=0.2)
    if save:
        matplot2tikz.save(f'jiont_isi.tex', encoding='utf8', axis_width='7.5cm', axis_height='5cm',
                          standalone=True)
        plt.savefig(f'joint_isi.pdf')
    else:
        plt.show()


def plot_rank_R_results_for_paper(folder1, folder2, n_montecarlo, save=False):
    results_differentR1 = np.load(Path(Path(__file__).parent.parent, f'simulation_results/{folder1}/different_R.npy'),
                                  allow_pickle=True).item()
    results_differentR2 = np.load(Path(Path(__file__).parent.parent, f'simulation_results/{folder2}/different_R.npy'),
                                  allow_pickle=True).item()

    # store R value results for each algorithm
    R_values = [1, 2, 5, 10, 20, 50]
    scenarios_differentR = [f'rank_{R}' for R in R_values]

    scenario_labels_differentR = [f'${R}$' for R in R_values]
    n_scenarios_differentR = len(scenario_labels_differentR)
    algorithms = list(results_differentR1[scenarios_differentR[0]].keys())
    joint_isi_per_algorithm_differentR1 = {algorithm: np.zeros((n_scenarios_differentR, n_montecarlo)) for algorithm in
                                           algorithms}
    joint_isi_per_algorithm_differentR2 = {algorithm: np.zeros((n_scenarios_differentR, n_montecarlo)) for algorithm in
                                           algorithms}

    for scenario_idx, scenario in enumerate(scenarios_differentR):
        for algorithm_idx, algorithm in enumerate(algorithms):
            joint_isi_per_algorithm_differentR1[algorithm][scenario_idx, :] = results_differentR1[scenario][algorithm][
                'joint_isi']
            joint_isi_per_algorithm_differentR2[algorithm][scenario_idx, :] = results_differentR2[scenario][algorithm][
                'joint_isi']

    # plot joint ISI for different R using true and estimated covariance matrices in one figure
    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(9, 2.5))

    colors = ['C0', 'C3', 'C2', 'C1', 'C4']
    # true
    axes[0].axhline(y=0.05, color='tab:gray', linestyle=':', linewidth=1.1)
    for (idx, algorithm) in enumerate(algorithms):
        axes[0].errorbar(R_values,
                         np.mean(joint_isi_per_algorithm_differentR1[algorithm], axis=1),
                         np.std(joint_isi_per_algorithm_differentR1[algorithm], axis=1),
                         color=colors[idx], linestyle=':', fmt='D', markersize=3, capsize=2, lw=1.1,
                         label=f'{algorithm}')
    axes[0].set_xscale('log')
    axes[0].set_xticks(R_values, scenario_labels_differentR, fontsize=12)
    axes[0].set_xlabel(r'effective rank $R$ \\ \\ (a) True covariance matrices', fontsize=12)
    axes[0].set_ylim([-0.05, 1.05])
    axes[0].set_yticks([0, 0.5, 1])
    axes[0].set_yticklabels([0, 0.5, 1], fontsize=12)
    axes[0].set_ylabel('jISI', fontsize=12)

    # estimated
    axes[1].axhline(y=0.05, color='tab:gray', linestyle=':', linewidth=1.1)
    for (idx, algorithm) in enumerate(algorithms):
        axes[1].errorbar(R_values,
                         np.mean(joint_isi_per_algorithm_differentR2[algorithm], axis=1),
                         np.std(joint_isi_per_algorithm_differentR2[algorithm], axis=1),
                         color=colors[idx], linestyle=':', fmt='D', markersize=3, capsize=2, lw=1.1,
                         label=f'{algorithm}')
    axes[1].set_xscale('log')
    axes[1].set_xticks(R_values, scenario_labels_differentR, fontsize=12)
    axes[1].set_xlabel(r'effective rank $R$ \\ \\ (b) Estimated covariance matrices', fontsize=12)
    axes[1].set_ylim([-0.05, 1.05])
    axes[1].set_yticks([0, 0.5, 1])
    axes[1].set_yticklabels([0, 0.5, 1], fontsize=12)

    plt.legend(loc='center left', bbox_to_anchor=(1.05, 0.5))

    if save:
        plt.tight_layout()
        plt.subplots_adjust(wspace=0.2)
        matplot2tikz.save('joint_isi_R.tex', encoding='utf8', axis_width='7.5cm', axis_height='5cm', standalone=True)
        plt.savefig(f'joint_isi_R.pdf')
    else:
        plt.title(f'joint ISI for the different experiments')
        plt.tight_layout()
        plt.subplots_adjust(wspace=0.2)
        plt.show()


def plot_different_samples_results_for_paper(folder, V_values, n_montecarlo, save=False):
    results_violations = {}
    for V in V_values:
        results_violations[V] = np.load(
            Path(Path(__file__).parent.parent, f'simulation_results/{folder}{V}/violations.npy'),
            allow_pickle=True).item()

    # store different samples results for each algorithm
    experiments = ['different_lambda_max', 'different_lambda_min']

    n_scenarios_violations = len(V_values)
    algorithms = list(results_violations[V_values[0]][experiments[0]].keys())

    joint_isi_per_algorithm_violations_expC = {algorithm: np.zeros((n_scenarios_violations, n_montecarlo)) for algorithm
                                               in algorithms}
    joint_isi_per_algorithm_violations_expD = {algorithm: np.zeros((n_scenarios_violations, n_montecarlo)) for algorithm
                                               in algorithms}

    for V_idx, V in enumerate(V_values):
        for algorithm_idx, algorithm in enumerate(algorithms):
            joint_isi_per_algorithm_violations_expC[algorithm][V_idx, :] = \
                results_violations[V]['different_lambda_min'][algorithm]['joint_isi']
            joint_isi_per_algorithm_violations_expD[algorithm][V_idx, :] = \
                results_violations[V]['different_lambda_max'][algorithm]['joint_isi']

    # label
    V_labels = [100, 1000, 10000, 100000, 1000000]
    V_label_values = [f'${V_value}$' for V_value in V_labels]

    # plot JISI for violations using true and estimated covariance matrices in one figure
    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(9, 2.5))

    colors = ['C0', 'C3', 'C2', 'C1', 'C4']

    # Experiment C
    axes[0].axhline(y=0.05, color='tab:gray', linestyle=':', linewidth=1.1)
    for (idx, algorithm) in enumerate(algorithms):
        axes[0].errorbar(V_values,
                         np.mean(joint_isi_per_algorithm_violations_expC[algorithm], axis=1),
                         np.std(joint_isi_per_algorithm_violations_expC[algorithm], axis=1),
                         color=colors[idx], linestyle=':', fmt='D', markersize=3, capsize=2, lw=1.1,
                         label=f'{algorithm}')
    axes[0].set_xscale('log')
    axes[0].set_xticks(V_labels, V_label_values, fontsize=12)
    axes[0].set_xlabel(r'number of samples $V$ \\ \\ (a) Experiment C', fontsize=12)
    axes[0].set_ylim([-0.05, 1.05])
    axes[0].set_yticks([0, 0.5, 1])
    axes[0].set_yticklabels([0, 0.5, 1], fontsize=12)
    axes[0].set_ylabel('jISI', fontsize=12)

    # Experiment D
    axes[1].axhline(y=0.05, color='tab:gray', linestyle=':', linewidth=1.1)
    for (idx, algorithm) in enumerate(algorithms):
        axes[1].errorbar(V_values,
                         np.mean(joint_isi_per_algorithm_violations_expD[algorithm], axis=1),
                         np.std(joint_isi_per_algorithm_violations_expD[algorithm], axis=1),
                         color=colors[idx], linestyle=':', fmt='D', markersize=3, capsize=2, lw=1.1,
                         label=f'{algorithm}')
    axes[1].set_xscale('log')
    axes[1].set_xticks(V_labels, V_label_values, fontsize=12)
    axes[1].set_xlabel(r'number of samples $V$ \\ \\ (b) Experiment D', fontsize=12)
    axes[1].set_ylim([-0.05, 1.05])
    axes[1].set_yticks([0, 0.5, 1])
    axes[1].set_yticklabels([0, 0.5, 1], fontsize=12)

    plt.legend(loc='center left', bbox_to_anchor=(1.05, 0.5))

    if save:
        plt.tight_layout()
        plt.subplots_adjust(wspace=0.2)
        matplot2tikz.save('joint_isi_samples.tex', encoding='utf8', axis_width='7.5cm', axis_height='5cm',
                          standalone=True)
        plt.savefig(f'joint_isi_samples.pdf')
    else:
        plt.title(f'joint ISI for the different experiments')
        plt.tight_layout()
        plt.subplots_adjust(wspace=0.2)
        plt.show()


def plot_all_eigenvalues_for_paper(scv_cov1, scv_cov2, scv_cov3, scv_cov4, filename=None):
    indices = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 99]
    Lambda = []
    Lambda1 = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov1)
    Lambda.append(Lambda1[:, ::-1][:, indices])  # sort descending
    Lambda2 = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov2)
    Lambda.append(Lambda2[:, ::-1][:, indices])  # sort descending
    Lambda3 = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov3)
    Lambda.append(Lambda3[:, ::-1][:, indices])  # sort descending
    Lambda4 = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov4)
    Lambda.append(Lambda4[:, ::-1][:, indices])  # sort descending
    titles = ['\large{Experiment A.} \\ \large{same ' + r'$\boldsymbol{\lambda}_n$, same $\boldsymbol{\Theta}_n$}',
              'Experiment B.' + r'\\' + r'same $\boldsymbol{\lambda}_n$, different $\boldsymbol{\Theta}_n$',
              'Experiment C.' + r'\\' + r'different $\lambda_n^{(\mathrm{min})}$',
              'Experiment D.' + r'\\' + r'different $\lambda_n^{(\mathrm{max})}$']

    fig, axes = plt.subplots(nrows=1, ncols=4, figsize=(10, 2.2))
    for ax_idx, ax in enumerate(axes):

        for n in range(scv_cov1.shape[2]):
            ax.plot(np.array(indices) + 1, Lambda[ax_idx][4 - n, :], 'D:', markersize=3, lw=1.5,
                    color=f'C{4 - n}', label=r'$\boldsymbol{\lambda}_' + f'{4 - n + 1}' + r'$')
        ax.set_xlabel(titles[ax_idx], fontsize=12)
        ax.set_xticks(np.array([0, 50, 100]), [0, 50, 100], fontsize=11)

    axes[0].set_ylim([-0.5, 5.5])
    axes[0].set_yticks(np.array([0, 2.5, 5]), [0, 2.5, 5], fontsize=11)

    axes[1].set_ylim([-0.5, 5.5])
    axes[1].set_yticks(np.array([0, 2.5, 5]), [0, 2.5, 5], fontsize=11)

    axes[2].set_ylim([-0.1, 1.1])
    axes[2].set_yticks(np.array([0, 0.5, 1]), [0, 0.5, 1], fontsize=11)

    axes[3].set_ylim([-5, 55])
    axes[3].set_yticks(np.array([0, 25, 50]), [0, 25, 50], fontsize=11)

    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    plt.tight_layout()
    plt.subplots_adjust(wspace=0.25)
    if filename is not None:
        matplot2tikz.save(f'{filename}.tex', encoding='utf8', axis_width='5.8cm', axis_height='4.4cm', standalone=True)
        plt.savefig(f'{filename}.pdf')
    else:
        plt.show()


def plot_all_eigenvalues_rank_R_for_paper(scv_cov1, scv_cov2, scv_cov3, scv_cov4, filename=None):
    indices = [0, 9, 19, 29, 39, 49, 59, 69, 79, 89, 99]
    Lambda = []
    Lambda1 = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov1)
    Lambda.append(Lambda1[:, ::-1][:, indices])  # sort descending
    Lambda2 = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov2)
    Lambda.append(Lambda2[:, ::-1][:, indices])  # sort descending
    Lambda3 = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov3)
    Lambda.append(Lambda3[:, ::-1][:, indices])  # sort descending
    Lambda4 = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov4)
    Lambda.append(Lambda4[:, ::-1][:, indices])  # sort descending
    titles = [r'\large{$R=1$}',
              r'\large{$R=10$}',
              r'\large{$R=20$}',
              r'\large{$R=50$}'
              ]

    fig, axes = plt.subplots(nrows=1, ncols=4, figsize=(10, 2.2))
    for ax_idx, ax in enumerate(axes):

        for n in range(scv_cov1.shape[2]):
            ax.plot(np.array(indices) + 1, Lambda[ax_idx][4 - n, :], 'D:', markersize=3.4, lw=2,
                    color=f'C{4 - n}', label=r'$\boldsymbol{\lambda}_' + f'{4 - n + 1}' + r'$')
        ax.set_xlabel(titles[ax_idx], fontsize=12)
        ax.set_xticks(np.array([0, 50, 100]), [0, 50, 100], fontsize=11)

    axes[0].set_ylim([-10, 110])
    axes[0].set_yticks(np.array([0, 50, 100]), [0, 50, 100], fontsize=11)

    axes[1].set_ylim([-2, 22])
    axes[1].set_yticks(np.array([0, 10, 20]), [0, 10, 20], fontsize=11)
    # #
    axes[2].set_ylim([-1, 11])
    axes[2].set_yticks(np.array([0, 5, 10]), [0, 5, 10], fontsize=11)
    #
    axes[3].set_ylim([-0.5, 5.5])
    axes[3].set_yticks(np.array([0, 2.5, 5]), [0, 2.5, 5], fontsize=11)

    plt.legend(loc='center left', bbox_to_anchor=(1.05, 0.5))
    plt.tight_layout()
    plt.subplots_adjust(wspace=0.25)
    if filename is not None:
        matplot2tikz.save(f'{filename}.tex', encoding='utf8', axis_width='5.8cm', axis_height='4.4cm', standalone=True)
        plt.savefig(f'{filename}.pdf')
    else:
        plt.show()



import numpy as np
from scipy.stats import random_correlation
from scipy.linalg import sqrtm, block_diag
from pathlib import Path

from independent_vector_analysis.helpers_iva import _bss_isi
from independent_vector_analysis.data_generation import MGGD_generation

from .helpers import calculate_eigenvalues_from_ccv_covariance_matrices
from .mcca import mcca

import time


def generate_datasets_from_covariance_matrices(scv_cov, V, orthogonal_A=False):
    # generate sources
    K, _, N = scv_cov.shape
    S = np.zeros((N, V, K))
    for n in range(N):
        S_temp = MGGD_generation(V, cov=scv_cov[:, :, n])[0]
        # make sources zero-mean and unit-variance
        S_temp -= np.mean(S_temp, axis=1, keepdims=True)
        S_temp /= np.std(S_temp, axis=1, keepdims=True)
        S[n, :, :] = S_temp.T

    # generate mixing matrices
    A = np.random.randn(N, N, K)
    if orthogonal_A:
        # make sure A is orthogonal -> X will be white for infinite samples
        for k in range(K):
            A[:, :, k] = np.linalg.solve(sqrtm(A[:, :, k] @ A[:, :, k].T), A[:, :, k])

    # generate observed datasets
    X = np.zeros((N, V, K))
    for k in range(K):
        X[:, :, k] = A[:, :, k] @ S[:, :, k]

    return X, A, S


def scv_covs_with_same_eigenvalues_same_eigenvectors_rank_K(N, K, alpha, beta):
    """
    Return SCV covariance matrices of dimensions (K,K,N) that have the same eigenvalues but different eigenvectors (to
    test if genvar can still identify the sources thanks to the eigenvectors).


    Parameters
    ----------
    N : int
        number of SCVs

    K : int
        number of datasets

    Returns
    -------
    scv_cov : np.ndarray
        Array of dimensions (K, K, N) that contains the SCV covariance matrices

    """

    # create N random SCV covariance matrices
    scv_cov = np.zeros((K, K, N))

    # same eigenvalue profile for all SCVs (generated as for rank K)
    scv_cov[:, :, 0] = scv_covs_with_rank_R(N, K, K, alpha, beta)[:, :, 0]
    Lambda, eigvec = np.linalg.eigh(scv_cov[:, :, 0])

    for n in range(1, N):
        # make matrix with diagonal elements +-1
        diag_elements = np.random.uniform(-1, 1, K)
        diag_elements = np.sign(diag_elements)
        D = np.diag(diag_elements)
        U = D @ eigvec
        scv_cov[:, :, n] = U @ np.diag(Lambda) @ U.T

    return scv_cov


def scv_covs_with_same_eigenvalues_different_eigenvectors_rank_K(N, K, alpha, beta):
    """
    Return SCV covariance matrices of dimensions (K,K,N) that have the same eigenvalues but different eigenvectors (to
    test if genvar can still identify the sources thanks to the eigenvectors).


    Parameters
    ----------
    N : int
        number of SCVs

    K : int
        number of datasets

    Returns
    -------
    scv_cov : np.ndarray
        Array of dimensions (K, K, N) that contains the SCV covariance matrices

    """

    # same eigenvalue profile for all SCVs (generated as for rank K)
    Lambda = calculate_eigenvalues_from_ccv_covariance_matrices(scv_covs_with_rank_R(N, K, K, alpha, beta))[0, :]

    # create N random SCV covariance matrices
    scv_cov = np.zeros((K, K, N))
    for n in range(N):
        scv_cov[:, :, n] = random_correlation.rvs(Lambda)

    return scv_cov


def scv_covs_with_rank_R(N, K, R, alpha, beta):
    """
    Return SCV covariance matrices of dimension (K,K,N) that are generated as
    C = alpha Q Q^T + beta G G^T + (1-alpha-beta) I, where Q is of dimensions (K,R) and G is of dimensions (K,K)


    Parameters
    ----------
    N : int
        number of SCVs

    K : int
        number of datasets

    R : int
        low rank of the model

    Returns
    -------
    scv_cov : np.ndarray
        Array of dimensions (K, K, N) that contains the SCV covariance matrices

    """
    while True:  # make sure that second-largest EVs of all SCVs are smaller than largest EVs of all SCVs
        scv_cov = np.zeros((K, K, N))
        for n in range(N):
            if alpha[n] + beta > 1:
                raise ValueError("alpha + beta must be smaller or equal to 1")
            temp_rank_term = np.random.randn(K, R)
            temp_rank_term /= np.linalg.norm(temp_rank_term, axis=1, keepdims=True)
            temp_variability_term = np.random.randn(K, K)
            temp_variability_term /= np.linalg.norm(temp_variability_term, axis=1, keepdims=True)
            scv_cov[:, :, n] = alpha[n] * (temp_rank_term @ temp_rank_term.T) + beta * (
                    temp_variability_term @ temp_variability_term.T) + (1 - alpha[n] - beta) * np.eye(K)

        # we assume that for R=1, if alpha is chosen properly, this does not need to be checked
        # (as for violating sumcor, the code would not run if this was tested)
        if R > 1:
            Lambda = calculate_eigenvalues_from_ccv_covariance_matrices(scv_cov)
            Lambda = Lambda[:, ::-1]  # sort descending

            # largest EVs of all SCVs should be bigger than second largest SCVs + some margin, otherwise recreate
            if np.min(Lambda[:, 0]) > np.max(Lambda[:, 1]) + 1 / K:
                break
        else:
            break

    return scv_cov


def save_joint_isi_and_runtime_results(N, K, V, n_montecarlo, scenarios, use_true_C_xx):
    folder = f'K_{K}_V_{V}'
    if use_true_C_xx:
        folder += '_true_C'

    algorithms = ['sumcor', 'maxvar', 'minvar', 'genvar', 'ssqcor']

    for run in range(n_montecarlo):
        print(f'Start run {run}...')

        for scenario_idx, scenario in enumerate(scenarios):
            print(f'Simulations for {scenario}')

            if scenario == 'same_eigenvalues_same_eigenvectors':
                scv_cov = scv_covs_with_same_eigenvalues_same_eigenvectors_rank_K(N, K,
                                                                                  alpha=[1, 1, 1, 1, 1],
                                                                                  beta=0.0)
            elif scenario == 'same_eigenvalues_different_eigenvectors':
                scv_cov = scv_covs_with_same_eigenvalues_different_eigenvectors_rank_K(N, K,
                                                                                       alpha=[1, 1, 1, 1, 1],
                                                                                       beta=0.0)
            elif scenario == 'different_lambda_min':
                alpha = 1 - (K - np.array([0.1, 0.15, 0.2, 0.25, 0.3])) / (K - 1)
                scv_cov = scv_covs_with_rank_R(N, K, 1, alpha=alpha, beta=0.0)
            elif scenario == 'different_lambda_max':
                alpha = 1 - (K - np.array([10, 15, 20, 25, 30])) / (K - 1)
                scv_cov = scv_covs_with_rank_R(N, K, 1, alpha=alpha, beta=0.0)
            elif scenario[0:5] == 'rank_':
                scv_cov = scv_covs_with_rank_R(N, K, int(scenario[5:]), alpha=[0.9, 0.85, 0.8, 0.75, 0.7], beta=0.0)
            else:
                raise AssertionError(f"scenario '{scenario}' does not exist")

            X, A, S = generate_datasets_from_covariance_matrices(scv_cov, V)

            if use_true_C_xx:
                # true joint SCV covariance matrix
                joint_scv_cov = block_diag(*list(scv_cov.T))

                # make the permutation matrix
                P = np.zeros((N * K, N * K))
                for n in range(N):
                    for k in range(K):
                        P[n + k * N, n * K + k] = 1

                # generate C_xx from true C_ss
                C_ss = P @ joint_scv_cov @ P.T
                A_joint = block_diag(*list(A.T)).T
                C_xx = A_joint @ C_ss @ A_joint.T
            else:
                C_xx = None

            for algorithm_idx, algorithm in enumerate(algorithms):
                t_start = time.process_time()
                M = mcca(X, algorithm, C_xx=C_xx)[0]
                W = np.moveaxis(M, [0, 1, 2], [1, 0, 2])
                t_end = time.process_time()

                filename = Path(Path(__file__).parent.parent,
                                f'simulation_results/{folder}/{scenario}_{algorithm}_run{run}.npy')
                np.save(filename, {'joint_isi': _bss_isi(W, A)[1], 'runtime': t_end - t_start})


def save_violation_results_from_multiple_files_in_one_file(folder, scenarios, n_montecarlo):
    algorithms = ['sumcor', 'maxvar', 'minvar', 'genvar', 'ssqcor']

    results = {}
    for scenario_idx, scenario in enumerate(scenarios):
        results[scenario] = {}
        for algorithm_idx, algorithm in enumerate(algorithms):
            joint_isi = np.zeros(n_montecarlo)
            runtime = np.zeros(n_montecarlo)
            for run in range(n_montecarlo):
                filename = Path(Path(__file__).parent.parent,
                                f'simulation_results/{folder}/{scenario}_{algorithm}_run{run}.npy')
                results_tmp = np.load(filename, allow_pickle=True).item()
                runtime[run] = results_tmp['runtime']
                joint_isi[run] = results_tmp['joint_isi']

            results[scenario][algorithm] = {'joint_isi': joint_isi, 'runtime': runtime}

    print(f'Save run as simulation_results/{folder}/violations.npy.')
    np.save(Path(Path(__file__).parent.parent, f'simulation_results/{folder}/violations.npy'), results)


def save_different_R_results_from_multiple_files_in_one_file(folder, n_montecarlo):
    scenarios = [f'rank_{R}' for R in [1, 2, 5, 10, 20, 50]]

    algorithms = ['sumcor', 'maxvar', 'minvar', 'genvar', 'ssqcor']

    results = {}
    for scenario_idx, scenario in enumerate(scenarios):
        results[scenario] = {}
        for algorithm_idx, algorithm in enumerate(algorithms):
            joint_isi = np.zeros(n_montecarlo)
            runtime = np.zeros(n_montecarlo)
            for run in range(n_montecarlo):
                filename = Path(Path(__file__).parent.parent,
                                f'simulation_results/{folder}/{scenario}_{algorithm}_run{run}.npy')
                results_tmp = np.load(filename, allow_pickle=True).item()
                runtime[run] = results_tmp['runtime']
                joint_isi[run] = results_tmp['joint_isi']

            results[scenario][algorithm] = {'joint_isi': joint_isi, 'runtime': runtime}

    print(f'Save run as simulation_results/{folder}/different_R.npy.')
    np.save(Path(Path(__file__).parent.parent, f'simulation_results/{folder}/different_R.npy'), results)
